/*-
 * Copyright (C) @BABOLO  2003 Jan 15 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2003 Jan 15 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: jailupctl.c,v 1.7 2011/09/30 01:33:27 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <mife.h>
#include <err.h>
#include <babolo/parser.h>
#include <babolo/BLINflag.h>
#include "jailup.h"
#include "jailupctl.h"

static const char *exclusives[] = {"syslogd", "named", NULL};

int
exclude(u_char *name) {
    int i;
    for (i = 0; exclusives[i] && exclusives[i][0]; i++)
        if  (0 == strncmp((char*)name, exclusives[i], NAMELENMAX)) return(i + 1);
    return(0);
}

/*
 * exec ${STARTUPDIR}/<tofun>/<class> <chdir> hostname IPaddr uid gid ...
 *                                           |<tail>
 */
pid_t
exup(u_int32_t flags, pid_t pid, char *path, u_char **args) {
    int p, st, ex = EX_OK;
    char **parm;

    for (p = 0; p && *(args[p]); p++);
    if  (!(parm = malloc(sizeof(char*) * p))) {
        ifBLIN_QV1(flags) warnx("exup malloc #1 %d", sizeof(char*) * p);
        errno = ENOMEM;
        ex = EX_OSERR;
    } else {
        parm[0] = path;
        ifBLIN_QV3(flags) fprintf(stderr, "Exup=%s~\n", path);
        for (p = 1; args[p + 1] && *(args[p + 1]); p++) {
            parm[p] = (char*)args[p + 1];
            ifBLIN_QV4(flags) fprintf(stderr, "[%d]=%s~\n", p, parm[p]);
        }
        parm[p + 1] = NULL;
        if  (!(NORUN & flags)) {
            if  (pid && waitpid(pid, &st, 0) < 0) ifBLIN_QV2(flags) warn("waitpid");
            switch (pid = fork()) {
            case -1: ifBLIN_QV1(flags) warn("fork");
                     ex = EX_OSERR;
                     break;
            case  0:/*
                     *if  (0 > daemon(0,1))                                        err(EX_OSERR, NULL);
                     */
                     execv(path, parm);
                     ifBLIN_QV1(flags) warn("execv");
                     _exit(EX_OSERR);
            default:  break;
        }   }
        free(parm);
    }
    ifBLIN_QV4(flags) fprintf(stderr,"Exup %d\n",ex);
    return(ex ? -1 : pid);
}

void
usage(int ex) {
    int c;

    fprintf( stderr
           , "jailupctl @BABOLO V.M "VERS"  "DATE"\n"
             "   Usage:\n"
             "jailupctl [-v[v[v]]] {start|restart|stop|restop|faststart} [genclass [dirjail]]\n"
             "   -v verbose\n"
             "   reads JAILUPRC file string by string by babolo_getparm()\n"
             "   and starts daemons STARTUPDIR/{start|restart|stop|restop}/{inclass} in {dirjail}\n"
             "   PREFIX, JAILUPRC, STARTUPDIR environment variables used\n"
             "Exclusives:\n"
           );
    for (c = 0; exclusives[c] && exclusives[c][0]; c++)
        fprintf(stderr, "    %d %s\n", c + 1, exclusives[c]);
    exit(ex);
}

int
main(int argc, char **argv, char **envp) {
    char *prefix = NULL, *startupdir = NULL, *jailuprc = NULL;
    char *todo, *genclass = NULL, *dirjail = NULL;
    char *(path[jsmax]) = {NULL, NULL, NULL, NULL};
    u_int32_t flocal, flags = BLIN_VER1;
#   define EFLAGNOW  1
#   define EFLAGNEXT (EFLAGNOW << 1)
#   define STONL     0x2000
    mife_descriptor *md = NULL;
    babolo_opt *opt = NULL;
    unsigned doma, dodoma;
    u_char *in = NULL, *p;
    int c, ex = EX_OK;
    u_char **sparced;
    babolo_parm *bp;
    pid_t pid;

    if  (babolo_openopt(Bpars_NONU | BLIN_VER1, &opt, argc, argv, "d:fhj:np:qv?")) {
        errx(EX_SOFTWARE, "babolo_openopt");
    }
    for (flocal = 0; (c = babolo_getopt(&opt)) != 0; flocal >>= 1) {
        babolo_parm *bp;
        char *o;

        switch (c) {
        case 'd': if  (!(startupdir = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(startupdir = getenv(startupdir))) usage(EX_USAGE);
                ; break;
        case 'f': flags |= FORCE;                                                          break;
        case 'j': if  (!(jailuprc = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(jailuprc = getenv(jailuprc))) usage(EX_USAGE);
                ; break;
        case 'n': flags |= NORUN;                                                          break;
        case 'p': if  (!(prefix = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(prefix = getenv(prefix))) usage(EX_USAGE);
                ; break;
        case 'q': flags &= ~BLIN_VERMASK;                                                  break;
        case 'r': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; bp = babolo_getparm( Bpars_NOEN | Bpars_NOAL | Bpars_NONU
                                     , (u_char**)&o
                                     , (u_char*)""
                                     , 0
                                     );
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 'R': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; if  ((flocal & EFLAGNOW) && !(o = getenv(o))) usage(EX_USAGE);
                ; bp = babolo_getparm(Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 'v': BLIN_VERBOSE(flags);                                                     break;
        case 'h':
        case '?': usage(EX_OK);                                                            break;
        default : usage(EX_USAGE);
    }   }
    do {
        if  (!prefix) prefix = getenv("PREFIX");
#       ifdef PREFIX
            if  (!prefix) prefix = PREFIX;
#       endif
        if  (!prefix) prefix = "/usr/local";

        if  (!startupdir) startupdir = getenv("STARTUPDIR");
#       ifdef STARTUPDIR
            if  (!startupdir) startupdir = STARTUPDIR;
#       else
            if  (!startupdir && !!prefix && (asprintf(&startupdir, "%s/etc/jailup", prefix) < 0)) {
                ifBLIN_QV1(flags) warn("asprintf #1");
                ex = EX_OSERR;
                break;
            }
#       endif

        if  (!jailuprc) jailuprc = getenv("JAILUPRC");
#       ifdef JAILUPRC
            if  (!jailuprc) jailuprc = JAILUPRC;
#       else
            if  (!jailuprc && !!startupdir && (asprintf(&jailuprc, "%s/rc", startupdir)) {
                ifBLIN_QV1(flags) warn("asprintf #2");
                ex = EX_OSERR;
                break;
            }
#       endif

        if  (!startupdir) {
            ifBLIN_QV1(flags) warnx("Not specified jailup startup dir");
            errno = EINVAL;
            ex = EX_USAGE;
            break;
        }
        if  (!jailuprc) {
            ifBLIN_QV1(flags) warnx("Not specified jailup config");
            errno = EINVAL;
            ex = EX_USAGE;
            break;
        }
        errno = 0;
        if  (!(todo = babolo_getarg(&opt))) {
            ifBLIN_QV1(flags) {
                if  (errno) warn("Error in command"); else warnx("No command");
            }
            errno = EINVAL;
            ex = EX_USAGE;
            break;
        }
        if  (!(genclass = babolo_getarg(&opt)) && errno) {
            ifBLIN_QV1(flags) warn("Error in genclass");
            errno = EINVAL;
            ex = EX_USAGE;
            break;
        }
        if  (!(dirjail = babolo_getarg(&opt)) && errno) {
            ifBLIN_QV1(flags) warn("Error in dirjail");
            errno = EINVAL;
            ex = EX_USAGE;
            break;
        }
        ifBLIN_QV3(flags)
            fprintf(stderr, "PREFIX=%s\nJAILUPRC=%s\nSTARTUPDIR=%s\n", prefix, jailuprc, startupdir);
        ifBLIN_QV4(flags) {
            fprintf(stderr, "Exclusives:\n");
            for (c = 0; exclusives[c] && exclusives[c][0]; c++)
                fprintf(stderr, "    %d %s\n", c + 1, exclusives[c]);
        }
        if  ((doma = babolo_testword(&jailupctl, (const u_char *)todo)) == 0) {
            ifBLIN_QV1(flags) warnx("Invalid command=%s~", todo);
            errno = EINVAL;
            ex = EX_USAGE;
            break;
        }
        ifBLIN_QV3(flags) fprintf(stderr, "Todo %s %X\n", todo, doma);
        if  (!(md = mife_open(MIFE_FULL, jailuprc))) {
            ifBLIN_QV1(flags) warn("mife_open");
            ex = EX_IOERR;
            break;
        }
        if  (mife_eolinit(md, "#") < 0) {
            ifBLIN_QV1(flags) warn("mife_eolinit");
            ex = EX_SOFTWARE;
            break;
        }
        if  (mife_read(md, 0, 0) < 0) {
            ifBLIN_QV1(flags) warn("mife_read");
            ex = EX_IOERR;
            break;
        }
        if  (!(in = mife_get(md, 0))) {
            ifBLIN_QV1(flags) warn("mife_get");
            ex = EX_IOERR;
            break;
        }
    } while(0);
    for (p = in; !ex && *p;) {
        flocal = 0;
        dodoma = doma;
        if  (!(bp = babolo_getparm(Bpars_FIXB | Bpars_LFLN, &p, (u_char*)"#", 0))) {
            ifBLIN_QV1(flags) warn("babolo_getparm");
            ex = EX_DATAERR;
            break;
        }
        if  (!(bp->argc)) continue;
        ifBLIN_QV5(flags) {
            fprintf(stderr, "=%d:\n", bp->argc);
            for (c = 0; c < bp->argc; c++) fprintf(stderr, "[%d]=%s~\n", c, bp->argv[c]);
        }
        if  (*(bp->argv[0])) {
            sparced = &(bp->argv[0]);
        } else {
            flocal |= STONL;
            sparced = &(bp->argv[1]);
        }
        ifBLIN_QV6(flags) for (c = 0; sparced[c]; c++) fprintf(stderr, "[%d]=%s~\n", c, sparced[c]);
        ifBLIN_QV3(flags) {
            fprintf( stderr, "%s genclass=%s, inclass=%s, dirjail=%s, ...=%s exclude=%d\n"
                   , (STONL & flocal) ? "Off" : "On "
                   , sparced[0], sparced[1], sparced[2], sparced[3]
                   , exclude(sparced[0])
                   );
            if  (genclass) fprintf( stderr, " genclass(%s==%s)=%s"
                                  , sparced[0], genclass
                                  , strcmp((char*)(sparced[0]), genclass) ? "No " : "Yes"
                                  );
            if  (dirjail) fprintf( stderr, " dirjail(%s==%s)=%s"
                                 , sparced[2], dirjail
                                 , strcmp((char*)(sparced[2]), dirjail) ? "No " : "Yes"
                                 );
            if  (genclass || dirjail) fprintf(stderr, "\n");
        }
        if  (genclass ? strcmp((char*)(sparced[0]), genclass) : exclude(sparced[0])) continue;
        if  (dirjail && strcmp((char*)(sparced[2]), dirjail)) continue;
        if  (STONL & flocal) dodoma = JSTOP | JRSTP;
        if  (JRSTR & dodoma) {
            if  (asprintf(&path[jrstr], "%s/restart/%s", startupdir, sparced[1]) < 0) {
                ifBLIN_QV1(flags) warn("asprintf #1 %s", "%s/restart/%s");
                if  (errno == ENOMEM) ex = EX_TEMPFAIL;
                else if (errno == EILSEQ) ex = EX_DATAERR;
                else ex = EX_OSERR;
                break;
            }
            if  (access(path[jrstr], X_OK) < 0) {
                free(path[jrstr]);
                path[jrstr] = NULL;
                dodoma &= ~JRSTR;
                dodoma |= NORSTR;
        }   }
        if  (JRSTP & dodoma) {
            if  (asprintf(&path[jrstp], "%s/restop/%s", startupdir, sparced[1]) < 0) {
                ifBLIN_QV1(flags) warn("asprintf #2 %s", "%s/restop/%s");
                if  (errno == ENOMEM) ex = EX_TEMPFAIL;
                else if (errno == EILSEQ) ex = EX_DATAERR;
                else ex = EX_OSERR;
                break;
            }
            if  (access(path[jrstp], X_OK) < 0) {
                free(path[jrstp]);
                path[jrstp] = NULL;
                dodoma &= ~JRSTP;
                dodoma |= NORSTP;
        }   }
        if  (JSTAR & dodoma) {
            if  (asprintf(&path[jstar], "%s/start/%s", startupdir, sparced[1]) < 0) {
                ifBLIN_QV1(flags) warn("asprintf #3 %s", "%s/start/%s");
                if  (errno == ENOMEM) ex = EX_TEMPFAIL;
                else if (errno == EILSEQ) ex = EX_DATAERR;
                else ex = EX_OSERR;
                break;
            }
            if  (access(path[jstar], X_OK) < 0) {
                ifBLIN_QV1(flags) warnx("Nothing to start");
                if  (!(flags & FORCE)) {
                    free(path[jstar]);
                    path[jstar] = NULL;
                    errno = ENOENT;
                    ex = EX_CONFIG;
                    break;
        }   }   }
        if  (JSTOP & dodoma) {
            if  (asprintf(&path[jstop], "%s/stop/%s", startupdir, sparced[1]) < 0) {
                ifBLIN_QV1(flags) warn("asprintf #3 %s", "%s/stop/%s");
                if  (errno == ENOMEM) ex = EX_TEMPFAIL;
                else if (errno == EILSEQ) ex = EX_DATAERR;
                else ex = EX_OSERR;
                break;
            }
            if  (access(path[jstop], X_OK) < 0) {
                ifBLIN_QV1(flags) warnx("Nothing to stop");
                if  (!(flags & FORCE)) {
                    free(path[jstop]);
                    path[jstop] = NULL;
                    errno = ENOENT;
                    ex = EX_CONFIG;
                    break;
        }   }   }
        ifBLIN_QV3(flags) fprintf(stderr, "Todo %X\n", dodoma);
        for (pid = 0, c = 0; c < jsmax; c++) {
            if  ((1 << c) & dodoma) {
                if  (setenv("GENCLASS", (char*)sparced[0], 1)) {
                    ifBLIN_QV2(flags) warn("Error in setenv GENCLASS=%s~", sparced[0]);
                }
                if  ((pid = exup(flags, pid, path[c], sparced)) < 0) {
                    ifBLIN_QV2(flags) warn( "exup %X %s %s %s %s"
                                          , (1 << c)
                                          , sparced[0]
                                          , sparced[1]
                                          , sparced[2]
                                          , sparced[3]
                                          );
            }   }
            free(path[c]);
            path[c] = NULL;
        }
        babolo_freeparm(bp);
    }
    for (c = jrstr; c < jsmax; c++) if  (path[c]) free(path[c]);
    if  (md) mife_close(md);
    return(ex);
}
